@@ -497,17 +497,17 @@ zfsetparam(char *name, void *val, int flags)
     int type = (flags & ZFPM_INTEGER) ? PM_INTEGER : PM_SCALAR;
 
     if (!(pm = (Param) paramtab->getnode(paramtab, name))
-	|| (pm->flags & PM_UNSET)) {
+	|| (pm->node.flags & PM_UNSET)) {
 	/*
 	 * just make it readonly when creating, in case user
 	 * *really* knows what they're doing
 	 */
 	if ((pm = createparam(name, type)) && (flags & ZFPM_READONLY))
-	    pm->flags |= PM_READONLY;
+	    pm->node.flags |= PM_READONLY;
     } else if (flags & ZFPM_IFUNSET) {
 	pm = NULL;
     }
-    if (!pm || PM_TYPE(pm->flags) != type) {
+    if (!pm || PM_TYPE(pm->node.flags) != type) {
 	/* parameters are funny, you just never know */
 	if (type == PM_SCALAR)
 	    zsfree((char *)val);
@@ -531,7 +531,7 @@ zfunsetparam(char *name)
     Param pm;
 
     if ((pm = (Param) paramtab->getnode(paramtab, name))) {
-	pm->flags &= ~PM_READONLY;
+	pm->node.flags &= ~PM_READONLY;
 	unsetparam_pm(pm, 0, 1);
     }
 }
@@ -582,7 +582,7 @@ zfgetline(char *ln, int lnsize, int tmout)
 
     if (setjmp(zfalrmbuf)) {
 	alarm(0);
-	zwarnnam("zftp", "timeout getting response", NULL, 0);
+	zwarnnam("zftp", "timeout getting response");
 	return 6;
     }
     zfalarm(tmout);
@@ -792,7 +792,7 @@ zfgetmsg(void)
 	zcfinish = 2;		/* don't need to tell server */
 	zfclose(0);
 	/* unexpected, so tell user */
-	zwarnnam("zftp", "remote server has closed connection", NULL, 0);
+	zwarnnam("zftp", "remote server has closed connection");
 	return 6;
     }
     if (lastcode == 530) {
@@ -805,7 +805,7 @@ zfgetmsg(void)
      * It means we just hang around waiting for another reply.
      */
     if (lastcode == 120) {
-	zwarnnam("zftp", "delay expected, waiting: %s", lastmsg, 0);
+	zwarnnam("zftp", "delay expected, waiting: %s", lastmsg);
 	return zfgetmsg();
     }
 
@@ -836,7 +836,7 @@ zfsendcmd(char *cmd)
     tmout = getiparam("ZFTP_TMOUT");
     if (setjmp(zfalrmbuf)) {
 	alarm(0);
-	zwarnnam("zftp", "timeout sending message", NULL, 0);
+	zwarnnam("zftp", "timeout sending message");
 	return 6;
     }
     zfalarm(tmout);
@@ -844,8 +844,7 @@ zfsendcmd(char *cmd)
     alarm(0);
 
     if (ret <= 0) {
-	zwarnnam("zftp send", "failure sending control message: %e",
-		 NULL, errno);
+	zwarnnam("zftp send", "failure sending control message: %e", errno);
 	return 6;
     }
 
@@ -860,12 +859,12 @@ static int
 zfopendata(char *name, union tcp_sockaddr *zdsockp, int *is_passivep)
 {
     if (!(zfprefs & (ZFPF_SNDP|ZFPF_PASV))) {
-	zwarnnam(name, "Must set preference S or P to transfer data", NULL, 0);
+	zwarnnam(name, "Must set preference S or P to transfer data");
 	return 1;
     }
     zfsess->dfd = socket(zfsess->control->peer.a.sa_family, SOCK_STREAM, 0);
     if (zfsess->dfd < 0) {
-	zwarnnam(name, "can't get data socket: %e", NULL, errno);
+	zwarnnam(name, "can't get data socket: %e", errno);
 	return 1;
     }
 
@@ -900,7 +899,7 @@ zfopendata(char *name, union tcp_sockaddr *zdsockp, int *is_passivep)
 	    ptr = strchr(lastmsg, '(');
 	    if(!ptr) {
 	    bad_epsv:
-		zwarnnam(name, "bad response to EPSV: %s", lastmsg, 0);
+		zwarnnam(name, "bad response to EPSV: %s", lastmsg);
 		zfclosedata();
 		return 1;
 	    }
@@ -940,7 +939,7 @@ zfopendata(char *name, union tcp_sockaddr *zdsockp, int *is_passivep)
 		    break;
 	    if (sscanf(ptr, "%d,%d,%d,%d,%d,%d",
 		       nums, nums+1, nums+2, nums+3, nums+4, nums+5) != 6) {
-		zwarnnam(name, "bad response to PASV: %s", lastmsg, 0);
+		zwarnnam(name, "bad response to PASV: %s", lastmsg);
 		zfclosedata();
 		return 1;
 	    }
@@ -960,7 +959,7 @@ zfopendata(char *name, union tcp_sockaddr *zdsockp, int *is_passivep)
 	} while (err && errno == EINTR && !errflag);
 
 	if (err) {
-	    zwarnnam(name, "connect failed: %e", NULL, errno);
+	    zwarnnam(name, "connect failed: %e", errno);
 	    zfclosedata();
 	    return 1;
 	}
@@ -976,7 +975,7 @@ zfopendata(char *name, union tcp_sockaddr *zdsockp, int *is_passivep)
 	int ret;
 
 	if (!(zfprefs & ZFPF_SNDP)) {
-	    zwarnnam(name, "only sendport mode available for data", NULL, 0);
+	    zwarnnam(name, "only sendport mode available for data");
 	    return 1;
 	}
 
@@ -1027,7 +1026,7 @@ zfopendata(char *name, union tcp_sockaddr *zdsockp, int *is_passivep)
 		    addr[0],addr[1],addr[2],addr[3],port[0],port[1]);
 	}
 	if (zfsendcmd(portcmd) >= 5) {
-	    zwarnnam(name, "port command failed", NULL, 0);
+	    zwarnnam(name, "port command failed");
 	    zfclosedata();
 	    return 1;
 	}
@@ -1129,7 +1128,7 @@ zfgetdata(char *name, char *rest, char *cmd, int getsize)
 	newfd = zfmovefd(accept(zfsess->dfd, (struct sockaddr *)&zdsock, 
 				&len));
 	if (newfd < 0)
-	    zwarnnam(name, "unable to accept data: %e", NULL, errno);
+	    zwarnnam(name, "unable to accept data: %e", errno);
 	zfclosedata();
 	if (newfd < 0)
 	    return 1;
@@ -1320,7 +1319,7 @@ zfread(int fd, char *bf, off_t sz, int tmout)
 
     if (setjmp(zfalrmbuf)) {
 	alarm(0);
-	zwarnnam("zftp", "timeout on network read", NULL, 0);
+	zwarnnam("zftp", "timeout on network read");
 	return -1;
     }
     zfalarm(tmout);
@@ -1345,7 +1344,7 @@ zfwrite(int fd, char *bf, off_t sz, int tmout)
 
     if (setjmp(zfalrmbuf)) {
 	alarm(0);
-	zwarnnam("zftp", "timeout on network write", NULL, 0);
+	zwarnnam("zftp", "timeout on network write");
 	return -1;
     }
     zfalarm(tmout);
@@ -1375,7 +1374,7 @@ zfread_block(int fd, char *bf, off_t sz, int tmout)
 	    n = zfread(fd, (char *)&hdr, sizeof(hdr), tmout);
 	} while (n < 0 && errno == EINTR);
 	if (n != 3 && !zfdrrrring) {
-	    zwarnnam("zftp", "failure reading FTP block header", NULL, 0);
+	    zwarnnam("zftp", "failure reading FTP block header");
 	    return n;
 	}
 	/* size is stored in network byte order */
@@ -1386,7 +1385,7 @@ zfread_block(int fd, char *bf, off_t sz, int tmout)
 	    /*
 	     * See comments in file headers
 	     */
-	    zwarnnam("zftp", "block too large to handle", NULL, 0);
+	    zwarnnam("zftp", "block too large to handle");
 	    errno = EIO;
 	    return -1;
 	}
@@ -1403,7 +1402,7 @@ zfread_block(int fd, char *bf, off_t sz, int tmout)
 		break;
 	}
 	if (cnt) {
-	    zwarnnam("zftp", "short data block", NULL, 0);
+	    zwarnnam("zftp", "short data block");
 	    errno = EIO;
 	    return -1;
 	}
@@ -1429,7 +1428,7 @@ zfwrite_block(int fd, char *bf, off_t sz, int tmout)
 	n = zfwrite(fd, (char *)&hdr, sizeof(hdr), tmout);
     } while (n < 0 && errno == EINTR);
     if (n != 3 && !zfdrrrring) {
-	zwarnnam("zftp", "failure writing FTP block header", NULL, 0);
+	zwarnnam("zftp", "failure writing FTP block header");
 	return n;
     }
     bfptr = bf;
@@ -1470,9 +1469,9 @@ zfsenddata(char *name, int recv, int progress, off_t startat)
     char lsbuf[ZF_BUFSIZE], *ascbuf = NULL, *optr;
     off_t sofar = 0, last_sofar = 0;
     readwrite_t read_ptr = zfread, write_ptr = zfwrite;
-    Eprog prog;
+    Shfunc shfunc;
 
-    if (progress && (prog = getshfunc("zftp_progress")) != &dummy_eprog) {
+    if (progress && (shfunc = getshfunc("zftp_progress"))) {
 	/*
 	 * progress to set up:  ZFTP_COUNT is zero.
 	 * We do this here in case we needed to wait for a RETR
@@ -1481,7 +1480,7 @@ zfsenddata(char *name, int recv, int progress, off_t startat)
 	int osc = sfcontext;
 
 	sfcontext = SFC_HOOK;
-	doshfunc("zftp_progress", prog, NULL, 0, 1);
+	doshfunc(shfunc, NULL, 1);
 	sfcontext = osc;
 	/* Now add in the bit of the file we've got/sent already */
 	sofar = last_sofar = startat;
@@ -1586,7 +1585,7 @@ zfsenddata(char *name, int recv, int progress, off_t startat)
 			if (!zfdrrrring &&
 			    (!interact || (!errflag && errno != EPIPE))) {
 			    ret = recv ? 2 : 1;
-			    zwarnnam(name, "write failed: %e", NULL, errno);
+			    zwarnnam(name, "write failed: %e", errno);
 			} else
 			    ret = recv ? 3 : 1;
 			break;
@@ -1601,7 +1600,7 @@ zfsenddata(char *name, int recv, int progress, off_t startat)
 		if (!zfdrrrring &&
 		    (!interact || (!errflag && errno != EPIPE))) {
 		    ret = recv ? 1 : 2;
-		    zwarnnam(name, "read failed: %e", NULL, errno);
+		    zwarnnam(name, "read failed: %e", errno);
 		} else
 		    ret = recv ? 1 : 3;
 		break;
@@ -1609,12 +1608,12 @@ zfsenddata(char *name, int recv, int progress, off_t startat)
 	} else
 	    break;
 	if (!ret && sofar != last_sofar && progress &&
-	    (prog = getshfunc("zftp_progress")) != &dummy_eprog) {
+	    (shfunc = getshfunc("zftp_progress"))) {
 	    int osc = sfcontext;
 
 	    zfsetparam("ZFTP_COUNT", &sofar, ZFPM_READONLY|ZFPM_INTEGER);
 	    sfcontext = SFC_HOOK;
-	    doshfunc("zftp_progress", prog, NULL, 0, 1);
+	    doshfunc(shfunc, NULL, 1);
 	    sfcontext = osc;
 	    last_sofar = sofar;
 	}
@@ -1649,7 +1648,7 @@ zfsenddata(char *name, int recv, int progress, off_t startat)
 	unsigned char msg[4] = { IAC, IP, IAC, SYNCH };
 
 	if (ret == 2)
-	    zwarnnam(name, "aborting data transfer...", NULL, 0);
+	    zwarnnam(name, "aborting data transfer...");
 
 	holdintr();
 
@@ -1698,7 +1697,7 @@ zftp_open(char *name, char **args, int flags)
 {
     struct protoent *zprotop;
     struct servent *zservp;
-    struct hostent *zhostp = NULL;
+    struct hostent *zhostp;
     char **addrp, *fname, *tmpptr, *portnam = "ftp";
     char *hostnam, *hostsuffix;
     int err, tmout, port = -1;
@@ -1709,7 +1708,7 @@ zftp_open(char *name, char **args, int flags)
 	if (zfsess->userparams)
 	    args = zfsess->userparams;
 	else {
-	    zwarnnam(name, "no host specified", NULL, 0);
+	    zwarnnam(name, "no host specified");
 	    return 1;
 	}
     }
@@ -1731,7 +1730,7 @@ zftp_open(char *name, char **args, int flags)
 	hostnam++;
 	hostsuffix = strchr(hostnam, ']');
 	if (!hostsuffix || (hostsuffix[1] && hostsuffix[1] != ':')) {
-	    zwarnnam(name, "Invalid host format: %s", hostnam, 0);
+	    zwarnnam(name, "Invalid host format: %s", hostnam);
 	    return 1;
 	}
 	*hostsuffix++ = '\0';
@@ -1761,8 +1760,7 @@ zftp_open(char *name, char **args, int flags)
     /* this is going to give 0.  why bother? */
     zprotop = getprotobyname("tcp");
     if (!zprotop) {
-	zwarnnam(name, "Can't find protocol TCP (is your network functional)?",
-		 NULL, 0);
+	zwarnnam(name, "Can't find protocol TCP (is your network functional)?");
 	return 1;
     }
     if (port < 0)
@@ -1771,7 +1769,7 @@ zftp_open(char *name, char **args, int flags)
 	zservp = getservbyport(port, "tcp");
 
     if (!zprotop || !zservp) {
-	zwarnnam(name, "Can't find port for service `%s'", portnam, 0);
+	zwarnnam(name, "Can't find port for service `%s'", portnam);
 	return 1;
     }
 
@@ -1788,9 +1786,9 @@ zftp_open(char *name, char **args, int flags)
 	alarm(0);
 	queue_signals();
 	if ((hname = getsparam("ZFTP_HOST")) && *hname) 
-	    zwarnnam(name, "timeout connecting to %s", hname, 0);
+	    zwarnnam(name, "timeout connecting to %s", hname);
 	else
-	    zwarnnam(name, "timeout on host name lookup", NULL, 0);
+	    zwarnnam(name, "timeout on host name lookup");
 	unqueue_signals();
 	zfclose(0);
 	return 1;
@@ -1818,7 +1816,7 @@ zftp_open(char *name, char **args, int flags)
 	     * on the other hand, herror() is obsolete
 	     */
 	    FAILED();
-	    zwarnnam(name, "host not found: %s", hostnam, 0);
+	    zwarnnam(name, "host not found: %s", hostnam);
 	    alarm(0);
 	    return 1;
 	}
@@ -1851,7 +1849,7 @@ zftp_open(char *name, char **args, int flags)
 	    zfunsetparam("ZFTP_HOST");
 	    zfunsetparam("ZFTP_PORT");
 	    FAILED();
-	    zwarnnam(name, "socket failed: %e", NULL, errno);
+	    zwarnnam(name, "socket failed: %e", errno);
 	    alarm(0);
 	    return 1;
 	}
@@ -1868,9 +1866,10 @@ zftp_open(char *name, char **args, int flags)
 	/* try all possible IP's */
 	for (addrp = zhostp->h_addr_list; err && *addrp; addrp++) {
 	    if(hlen != zhostp->h_length)
-		zwarnnam(name, "address length mismatch", NULL, 0);
+		zwarnnam(name, "address length mismatch");
 	    do {
-		err = tcp_connect(zfsess->control, *addrp, zhostp, zservp->s_port);	    } while (err && errno == EINTR && !errflag);
+		err = tcp_connect(zfsess->control, *addrp, zhostp, zservp->s_port);
+	    } while (err && errno == EINTR && !errflag);
 	    /* you can check whether it's worth retrying here */
 	}
 
@@ -1878,7 +1877,7 @@ zftp_open(char *name, char **args, int flags)
 	    freehostent(zhostp);
 	    zfclose(0);
 	    FAILED();
-	    zwarnnam(name, "connect failed: %e", NULL, errno);
+	    zwarnnam(name, "connect failed: %e", errno);
 	    alarm(0);
 	    return 1;
 	}
@@ -1913,7 +1912,7 @@ zftp_open(char *name, char **args, int flags)
 
     len = sizeof(zfsess->control->sock);
     if (getsockname(zfsess->control->fd, (struct sockaddr *)&zfsess->control->sock, &len) < 0) {
-	zwarnnam(name, "getsockname failed: %e", NULL, errno);
+	zwarnnam(name, "getsockname failed: %e", errno);
 	zfclose(0);
 	return 1;
     }
@@ -1940,7 +1939,7 @@ zftp_open(char *name, char **args, int flags)
     zfsess->cin = fdopen(zfsess->control->fd, "r");
 
     if (!zfsess->cin) {
-	zwarnnam(name, "file handling error", NULL, 0);
+	zwarnnam(name, "file handling error");
 	zfclose(0);
 	return 1;
     }
@@ -2168,7 +2167,7 @@ zftp_login(char *name, char **args, UNUSED(int flags))
 	    else
 		acct = zfgetinfo("Account: ", 0);
 	    zsfree(ucmd);
-	    ucmd = tricat("ACCT ", passwd, "\r\n");
+	    ucmd = tricat("ACCT ", acct, "\r\n");
 	    if (zfsendcmd(ucmd) == 6)
 		stopit = 2;
 	    break;
@@ -2189,7 +2188,7 @@ zftp_login(char *name, char **args, UNUSED(int flags))
     if (!zfsess->control)
 	return 1;
     if (stopit == 2 || (lastcode != 230 && lastcode != 202)) {
-	zwarnnam(name, "login failed", NULL, 0);
+	zwarnnam(name, "login failed");
 	return 1;
     }
 
@@ -2197,7 +2196,7 @@ zftp_login(char *name, char **args, UNUSED(int flags))
 	int cnt;
 	for (cnt = 0; *args; args++)
 	    cnt++;
-	zwarnnam(name, "warning: %d comand arguments not used\n", NULL, cnt);
+	zwarnnam(name, "warning: %d comand arguments not used\n", cnt);
     }
     zfstatusp[zfsessno] |= ZFST_LOGI;
     zfsetparam("ZFTP_USER", ztrdup(user), ZFPM_READONLY);
@@ -2298,7 +2297,7 @@ zftp_test(UNUSED(char *name), UNUSED(char **args), UNUSED(int flags))
     /* if we have no zfsess->control, then we've just been dumped out. */
     return zfsess->control ? 0 : 2;
 #else
-    zfwarnnam(name, "not supported on this system.", NULL, 0);
+    zfwarnnam(name, "not supported on this system.");
     return 3;
 #endif /* defined(HAVE_POLL) || defined(HAVE_SELECT) */
 }
@@ -2365,7 +2364,7 @@ zfgetcwd(void)
 {
     char *ptr, *eptr;
     int endc;
-    Eprog prog;
+    Shfunc shfunc;
 
     if (zfprefs & ZFPF_DUMB)
 	return 1;
@@ -2392,11 +2391,11 @@ zfgetcwd(void)
      * front end.  By putting it here, and in close when ZFTP_PWD is unset,
      * we at least cover the bases.
      */
-    if ((prog = getshfunc("zftp_chpwd")) != &dummy_eprog) {
+    if ((shfunc = getshfunc("zftp_chpwd"))) {
 	int osc = sfcontext;
 
 	sfcontext = SFC_HOOK;
-	doshfunc("zftp_chpwd", prog, NULL, 0, 1);
+	doshfunc(shfunc, NULL, 1);
 	sfcontext = osc;
     }
     return 0;
@@ -2450,7 +2449,7 @@ zftp_type(char *name, char **args, int flags)
 	 * ones we know what to do with.
 	 */
 	if (str[1] || (nt != 'A' && nt != 'B' && nt != 'I')) {
-	    zwarnnam(name, "transfer type %s not recognised", str, 0);
+	    zwarnnam(name, "transfer type %s not recognised", str);
 	    return 1;
 	}
 	
@@ -2480,7 +2479,7 @@ zftp_mode(char *name, char **args, UNUSED(int flags))
     }
     nt = str[0] = toupper(STOUC(*str));
     if (str[1] || (nt != 'S' && nt != 'B')) {
-	zwarnnam(name, "transfer mode %s not recognised", str, 0);
+	zwarnnam(name, "transfer mode %s not recognised", str);
 	return 1;
     }
     cmd[5] = (char) nt;
@@ -2550,7 +2549,7 @@ zftp_getput(char *name, char **args, int flags)
 {
     int ret = 0, recv = (flags & ZFTP_RECV), getsize = 0, progress = 1;
     char *cmd = recv ? "RETR " : (flags & ZFTP_APPE) ? "APPE " : "STOR ";
-    Eprog prog;
+    Shfunc shfunc;
 
     /*
      * At this point I'd like to set progress to 0 if we're
@@ -2568,8 +2567,8 @@ zftp_getput(char *name, char **args, int flags)
     for (; *args; args++) {
 	char *ln, *rest = NULL;
 	off_t startat = 0;
-	if (progress && (prog = getshfunc("zftp_progress")) != &dummy_eprog) {
-	    off_t sz;
+	if (progress && (shfunc = getshfunc("zftp_progress"))) {
+	    off_t sz = -1;
 	    /*
 	     * This calls the SIZE command to get the size for remote
 	     * files.  Some servers send the size with the reply to
@@ -2609,14 +2608,14 @@ zftp_getput(char *name, char **args, int flags)
 	 * if and only if we called zfsenddata();
 	 */
 	if (progress && ret != 2 &&
-	    (prog = getshfunc("zftp_progress")) != &dummy_eprog) {
+	    (shfunc = getshfunc("zftp_progress"))) {
 	    /* progress to finish: ZFTP_TRANSFER set to GF or PF */
 	    int osc = sfcontext;
 
 	    zfsetparam("ZFTP_TRANSFER", ztrdup(recv ? "GF" : "PF"),
 		       ZFPM_READONLY);
 	    sfcontext = SFC_HOOK;
-	    doshfunc("zftp_progress", prog, NULL, 0, 1);
+	    doshfunc(shfunc, NULL, 1);
 	    sfcontext = osc;
 	}
 	if (rest) {
@@ -2716,7 +2715,7 @@ static void
 zfclose(int leaveparams)
 {
     char **aptr;
-    Eprog prog;
+    Shfunc shfunc;
 
     if (!zfsess->control)
 	return;
@@ -2767,11 +2766,11 @@ zfclose(int leaveparams)
 	    zfunsetparam(*aptr);
 
 	/* Now ZFTP_PWD is unset.  It's up to zftp_chpwd to notice. */
-	if ((prog = getshfunc("zftp_chpwd")) != &dummy_eprog) {
+	if ((shfunc = getshfunc("zftp_chpwd"))) {
 	    int osc = sfcontext;
 
 	    sfcontext = SFC_HOOK;
-	    doshfunc("zftp_chpwd", prog, NULL, 0, 1);
+	    doshfunc(shfunc, NULL, 1);
 	    sfcontext = osc;
 	}
     }
@@ -3016,7 +3015,7 @@ bin_zftp(char *name, char **args, UNUSED(Options ops), UNUSED(int func))
 	    break;
 
     if (!zptr->nam) {
-	zwarnnam(name, "no such subcommand: %s", cnam, 0);
+	zwarnnam(name, "no such subcommand: %s", cnam);
 	return 1;
     }
 
@@ -3024,7 +3023,7 @@ bin_zftp(char *name, char **args, UNUSED(Options ops), UNUSED(int func))
     for (n = 0; args[n]; n++)
 	;
     if (n < zptr->min || (zptr->max != -1 && n > zptr->max)) {
-	zwarnnam(name, "wrong no. of arguments for %s", cnam, 0);
+	zwarnnam(name, "wrong no. of arguments for %s", cnam);
 	return 1;
     }
 
@@ -3071,7 +3070,7 @@ bin_zftp(char *name, char **args, UNUSED(Options ops), UNUSED(int func))
 	     * with ret == 2, we just got dumped out in the test,
 	     * so enough messages already.
 	     */	       
-	    zwarnnam(fullname, "not connected.", NULL, 0);
+	    zwarnnam(fullname, "not connected.");
 	}
 	return 1;
     }
@@ -3102,7 +3101,7 @@ bin_zftp(char *name, char **args, UNUSED(Options ops), UNUSED(int func))
 		break;
 
 	    default:
-		zwarnnam(name, "preference %c not recognized", NULL, *ptr);
+		zwarnnam(name, "preference %c not recognized", *ptr);
 		break;
 	    }
 	}
@@ -3152,7 +3151,6 @@ zftp_cleanup(void)
     zfunsetparam("ZFTP_SESSION");
     freelinklist(zfsessions, (FreeFunc) freesession);
     zfree(zfstatusp, sizeof(int)*zfsesscnt);
-    deletebuiltins("zftp", bintab, sizeof(bintab)/sizeof(*bintab));
 }
 
 static int
@@ -3162,47 +3160,68 @@ zftpexithook(UNUSED(Hookdef d), UNUSED(void *dummy))
     return 0;
 }
 
+static struct features module_features = {
+    bintab, sizeof(bintab)/sizeof(*bintab),
+    NULL, 0,
+    NULL, 0,
+    NULL, 0,
+    0
+};
+
 /* The load/unload routines required by the zsh library interface */
 
 /**/
 int
 setup_(UNUSED(Module m))
 {
-    /* setup_ returns 0 for success. require_module returns 1 for success. */
-    return !require_module("", "zsh/net/tcp", 0, 0);
+    return (require_module("zsh/net/tcp", NULL) == 1);
+}
+
+/**/
+int
+features_(Module m, char ***features)
+{
+    *features = featuresarray(m, &module_features);
+    return 0;
+}
+
+/**/
+int
+enables_(Module m, int **enables)
+{
+    return handlefeatures(m, &module_features, enables);
 }
 
 /**/
 int
 boot_(UNUSED(Module m))
 {
-    int ret;
-    if ((ret = addbuiltins("zftp", bintab,
-			   sizeof(bintab)/sizeof(*bintab))) == 1) {
-	/* if successful, set some default parameters */
-	off_t tmout_def = 60;
-	zfsetparam("ZFTP_VERBOSE", ztrdup("450"), ZFPM_IFUNSET);
-	zfsetparam("ZFTP_TMOUT", &tmout_def, ZFPM_IFUNSET|ZFPM_INTEGER);
-	zfsetparam("ZFTP_PREFS", ztrdup("PS"), ZFPM_IFUNSET);
-	/* default preferences if user deletes variable */
-	zfprefs = ZFPF_SNDP|ZFPF_PASV;
+    /*
+     * Set some default parameters.
+     * These aren't special, so aren't associated with features.
+     */
+    off_t tmout_def = 60;
+    zfsetparam("ZFTP_VERBOSE", ztrdup("450"), ZFPM_IFUNSET);
+    zfsetparam("ZFTP_TMOUT", &tmout_def, ZFPM_IFUNSET|ZFPM_INTEGER);
+    zfsetparam("ZFTP_PREFS", ztrdup("PS"), ZFPM_IFUNSET);
+    /* default preferences if user deletes variable */
+    zfprefs = ZFPF_SNDP|ZFPF_PASV;
     
-	zfsessions = znewlinklist();
-	newsession("default");
+    zfsessions = znewlinklist();
+    newsession("default");
 
-	addhookfunc("exit", zftpexithook);
-    }
+    addhookfunc("exit", zftpexithook);
 
-    return !ret;
+    return 0;
 }
 
 /**/
 int
-cleanup_(UNUSED(Module m))
+cleanup_(Module m)
 {
     deletehookfunc("exit", zftpexithook);
     zftp_cleanup();
-    return 0;
+    return setfeatureenables(m, &module_features, NULL);
 }
 
 /**/
